import functools
import json

import osim_utils.exceptions as exc


def check_response(*args):
    """
    Checks that a call to an API client method returned an expected status code.
    Raises an AssertionError if the returned status code is unexpected.

    Args:
        *args (tuple): status codes that should NOT lead to an
            exception

    Returns:
        The response of the decorated method
    """

    def decorator(func):
        @functools.wraps(func)
        def wrapper(*inner_args, **kwargs):
            response = func(*inner_args, **kwargs)
            assert response.status_code in args, (
                f"API call initiated by {func.__module__}.{func.__name__} "
                f"returned error: {response.content}"
            )
            return response

        return wrapper

    return decorator


def process_response(func):
    """
    Processes the response of a call to an API client method
    so that only the body of the response is decoded form json and returned

    Returns:
        The json-decoded body of the response of the decorated method
    """

    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        response = func(*args, **kwargs)
        try:
            return response.json()
        except json.JSONDecodeError:
            raise exc.ApiClientError(
                f"API call initiated by {func.__module__}.{func.__name__} "
                f"failed with error: {response.text}"
            )

    return wrapper



import os
import requests_cache
from datetime import timedelta
from http import HTTPStatus
from typing import Optional, Union

from osim_utils.logger import get_logger
from osim_utils.decorators import check_response, process_response
from osim_utils.exceptions import DataNotFoundError, DataValidationError


class SherpaClient:
    """
    Client for JISC Sherpa APIs
    """

    def __init__(self, api_key: Optional[str] = None):
        self.base_endpoint = "https://v2.sherpa.ac.uk/cgi/retrieve"
        self.api_key = api_key
        if api_key is None:
            try:
                self.api_key = os.environ["SHERPA_KEY"]
            except KeyError:
                raise DataNotFoundError(
                    "Could not find API key for JISC Sherpa services; please store the key "
                    "value in environment variable SHERPA_KEY or pass the key value directly as attribute api_key"
                    "when instantiating this SherpaClient"
                )
        self.session = requests_cache.CachedSession(
            "sherpa_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()

    @process_response
    @check_response(HTTPStatus.OK)
    def get(self, **kwargs):
        params = {
            "api-key": self.api_key,
            "item-type": "publication",
            "format": "Json",
        } | kwargs.get("params", dict())
        r = self.session.get(url=self.base_endpoint, params=params)
        return r

    def get_journal_by_issn(self, issn: Union[str, list[str]]) -> dict:
        if isinstance(issn, str):
            params = {"filter": f'[["issn", "equals", "{issn}"]]'}
            r = self.get(params=params)
        elif isinstance(issn, list):
            params = {"filter": f'[["issn", "equals", "{issn[0]}"]]'}
            r = self.get(params=params)
            try:
                return r["items"][0]
            except IndexError:
                params = {"filter": f'[["issn", "equals", "{issn[1]}"]]'}
                r = self.get(params=params)
        else:
            raise DataValidationError(
                f"Attribute issn has to be a str or list; {type(issn)} found"
            )
        try:
            return r["items"][0]
        except IndexError:
            return dict()

    def get_journal_by_title(self, title: str) -> dict:
        params = {"filter": f'[["title", "equals", "{title}"]]'}
        r = self.get(params=params)
        try:
            return r["items"][0]
        except IndexError:
            return dict()



def chunk_list(lst: list, n: int) -> list[list]:
    """
    Breaks a list into sublists (chunks) of length n.
    Args:
        lst: the list to be broken into chunks
        n: the length of each chunk
    """
    return [lst[i : i + n] for i in range(0, len(lst), n)]



import os
import requests_cache
from datetime import timedelta
from http import HTTPStatus

import osim_utils.exceptions as exc
from osim_utils.decorators import check_response, process_response
from osim_utils.logger import get_logger


class DataWareClient:
    """
    API client for EMBL's data warehouse API (Converis mirror)
    """

    def __init__(self):
        self.base_url = "https://xs-db.embl.de/v2/publications/"
        self.session = requests_cache.CachedSession(
            "warehouse_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()

        token_env_var = "DataWareHouseToken"
        self.default_headers = dict()
        try:
            self.default_headers[
                "Authorization"
            ] = f"Bearer {os.environ[token_env_var]}"
        except KeyError:
            raise exc.AuthenticationError(
                f"Environment variable {token_env_var} not defined. If you do not have an authorization"
                f"token for the Converis mirror dataset in EMBL's Data Warehouse API, please contact"
                f"Converis administrators."
            )

    @process_response
    @check_response(HTTPStatus.OK)
    def _get_all_pubs(self, **kwargs):
        headers = self.default_headers | kwargs.get("headers", dict())
        return self.session.get(self.base_url, headers=headers, params={"all": "true"})

    def get_all_publications(self, **kwargs):
        r = self._get_all_pubs(**kwargs)
        unique_pubs = list()
        processed_pubs = list()
        for owner in r:
            pubs = owner["publications"]
            for k, v in pubs.items():
                if (pub_id := v["id"]) not in processed_pubs:
                    processed_pubs.append(pub_id)
                    unique_pubs.append(v)
        return unique_pubs



ror_id2site = {
    "https://ror.org/010jaxs89": "EMBL Barcelona",
    "https://ror.org/02catss52": "EMBL-EBI",
    "https://ror.org/01zjc6908": "EMBL Grenoble",
    "https://ror.org/050589e39": "EMBL Hamburg",
    "https://ror.org/03mstc592": "EMBL Heidelberg",
    "https://ror.org/01yr73893": "EMBL Rome",
}



import requests_cache
from datetime import timedelta
from http import HTTPStatus

from osim_utils.logger import get_logger
from osim_utils.decorators import check_response, process_response


class RorClient:
    """
    Client for ROR API
    """

    def __init__(self):
        self.base_endpoint = "https://api.ror.org/organizations"
        self.session = requests_cache.CachedSession(
            "ror_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()

    @process_response
    @check_response(HTTPStatus.OK)
    def get_organisation_by_id(self, ror_id: str) -> dict:
        """
        :param ror_id: supports both ROR id (04m01e293) and ROR url (https://ror.org/04m01e293)
        :return:
        """
        return self.session.get(url=f"{self.base_endpoint}/{ror_id.split('/')[-1]}")



import os
import requests_cache
from datetime import timedelta
from http import HTTPStatus
from typing import Optional

from osim_utils.logger import get_logger
from osim_utils.decorators import check_response, process_response


class CrossRefClient:
    """
    Client for CrossRef API
    """

    def __init__(self, user_email: Optional[str] = None):
        self.session = requests_cache.CachedSession(
            "crossref_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()
        self.user_email = user_email
        self.common_params = {"format": "json"}
        if user_email is None:
            try:
                self.user_email = os.environ["CROSSREF_EMAIL"]
            except KeyError:
                self.logger.warning(
                    f"Environment variable CROSSREF_EMAIL not defined. Requests will be made from the common pool"
                )

    @process_response
    @check_response(HTTPStatus.OK)
    def get(self, **kwargs):
        url = kwargs.pop("url")
        params = {
            "pid": self.user_email,
            "format": "json",
        } | kwargs.get("params", dict())
        r = self.session.get(url=url, params=params)
        return r

    def get_metadata_by_doi(self, doi):
        params = {"id": doi}
        return self.get(url="https://doi.crossref.org/servlet/query", params=params)



from typing import Optional


class DetailedError(Exception):
    def __init__(self, message: str, extra: Optional[dict] = None):
        """
        Args:
            message: Error message
            extra: Details for debugging (e.g. variable values, etc)
        """
        self.message = message
        self.extra = extra

    def __str__(self):
        return f"{self.message} --- {self.extra}"


class ImprovementNeededError(DetailedError):
    pass


class ApiClientError(DetailedError):
    pass


class AuthenticationError(DetailedError):
    pass


class DuplicateValueError(DetailedError):
    pass


class DataValidationError(DetailedError):
    pass


class DataMismatchError(DetailedError):
    pass


class DataNotFoundError(DetailedError):
    pass



import os
import requests_cache
import urllib.parse
from datetime import timedelta
from http import HTTPStatus
from typing import Optional

from osim_utils.logger import get_logger
from osim_utils.constants import ror_id2site
from osim_utils.decorators import check_response, process_response
from osim_utils.common import chunk_list


class OpenAlexClient:
    """
    Client for OpenAlex API
    """

    max_page_size = 50

    def __init__(self):
        self.base_endpoint = "https://api.openalex.org"
        self.works_endpoint = f"{self.base_endpoint}/works"
        self.session = requests_cache.CachedSession(
            "openalex_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()

        req_env_var = "OpenAlexRequester"
        self.default_headers = dict()
        try:
            self.default_headers["User-agent"] = f"mailto:{os.environ[req_env_var]}"
        except KeyError:
            self.logger.warning(
                f"Environment variable {req_env_var} not defined. Requests will be made from the common pool (see  "
                f"https://docs.openalex.org/how-to-use-the-api/rate-limits-and-authentication#the-polite-pool "
                f"for details)"
            )

    @process_response
    @check_response(HTTPStatus.OK)
    def get(self, url, **kwargs):
        headers = self.default_headers | kwargs.get("headers", dict())
        params = kwargs.get("params")
        return self.session.get(url, headers=headers, params=params)

    def get_work_by_doi(self, doi: str, **kwargs) -> dict:
        """
        Queries works by doi
        Args:
            doi: The doi of the publication we are querying for
            **kwargs: additional parameters to be passed to the OpenAlex API

        Returns: OpenAlex API response
        """
        return self.get(
            f"{self.works_endpoint}/doi:{urllib.parse.quote(doi)}", **kwargs
        )

    def get_works_by_doi_list(self, dois: list[str], **kwargs) -> dict:
        """
        Retrieves up to 50 works by querying for a list of dois
        Args:
            dois: list of dois to be queried
            **kwargs: additional parameters to be passed to the OpenAlex API

        Returns: OpenAlex API response
        """
        quoted_dois = [urllib.parse.quote(x) for x in dois]
        sublists = chunk_list(quoted_dois, self.max_page_size)
        for sl in sublists:
            query = "|".join(sl)
            params = {"filter": f"doi:{query}", "per-page": self.max_page_size}
            yield self.get(self.works_endpoint, params=params, **kwargs)

    def get_embl_works(self, filters: dict, **kwargs) -> dict:
        """
        Queries works associated with any of the EMBL ROR ids in ror_id2site

        Args:
            filters: additional filters to apply to query
                (see https://docs.openalex.org/how-to-use-the-api/get-lists-of-entities/filter-entity-lists)
            **kwargs: additional parameters to be passed to the OpenAlex API

        Returns:

        """
        filter_str = f"institutions.ror:{'|'.join(ror_id2site.keys())}"
        for k, v in filters.items():
            filter_str += f",{k}:{v}"
        params = {"filter": filter_str} | kwargs.pop("params", dict())
        return self.get(self.works_endpoint, params=params, **kwargs)

    def get_all_embl_works(self, filters: dict, **kwargs) -> dict:
        """
        Convenience method to get and return all results matching filter.
        In other words, this method loops through all pages of API results to return
        the total.

        Args:
            filters: additional filters to apply to query
                (see https://docs.openalex.org/how-to-use-the-api/get-lists-of-entities/filter-entity-lists)
            **kwargs: additional parameters to be passed to the OpenAlex API
        """
        cursor = "*"
        params = {"per-page": 200} | kwargs.pop("params", dict())
        results = list()
        while cursor:
            params.update(cursor=cursor)
            r = self.get_embl_works(filters=filters, params=params)
            results += r["results"]
            cursor = r["meta"]["next_cursor"]
        return {
            "meta": r["meta"],
            "results": results,
        }

    def get_works_by_orcids(
        self, orcids: list[str], filters: Optional[dict] = None, **kwargs
    ) -> dict:
        cursor = "*"
        filter_str = f"author.orcid:{'|'.join(orcids)}"
        if filters:
            for k, v in filters.items():
                filter_str += f",{k}:{v}"
        params = (
            {"filter": filter_str} | {"per-page": 200} | kwargs.pop("params", dict())
        )
        results = list()
        while cursor:
            params.update(cursor=cursor)
            r = self.get(self.works_endpoint, params=params, **kwargs)
            results += r["results"]
            cursor = r["meta"]["next_cursor"]
        return {
            "meta": r["meta"],
            "results": results,
        }



import requests_cache
from datetime import timedelta
from osim_utils.logger import get_logger

from http import HTTPStatus
from osim_utils.decorators import check_response, process_response


class EpmcClient:
    """
    API client for EuropePMC Articles REST API
    (https://europepmc.org/RestfulWebService)
    """

    def __init__(self):
        self.base_url = "https://www.ebi.ac.uk/europepmc/webservices/rest/"
        self.common_params = {"format": "json"}
        self.session = requests_cache.CachedSession(
            "epmc_cache", expire_after=timedelta(days=7)
        )
        self.logger = get_logger()

    @staticmethod
    def _check_query(**kwargs) -> str:
        """
        Check that required "query" argument was passed to calling method
        Args:
            **kwargs: calling method kwargs

        Returns: The query string
        """
        try:
            return kwargs["query"]
        except KeyError:
            raise ValueError(
                "You must include a 'query' parameter when calling this method"
            )

    @process_response
    @check_response(HTTPStatus.OK)
    def search(self, **kwargs) -> requests.models.Response:
        """
        Queries the publication database using the GET search endpoint

        Args:
            **kwargs:
                query (required; str): query string

        Returns: server's response to the search request
        """
        self._check_query(**kwargs)
        endpoint = f"{self.base_url}search"
        params = self.common_params | kwargs
        return self.session.get(endpoint, params=params)

    @process_response
    @check_response(HTTPStatus.OK)
    def search_post(self, **kwargs) -> requests.models.Response:
        """
        Queries the publication database using the POST searchPOST endpoint

        Args:
            **kwargs:
                query (str): query string
                resultType (str): It can have the following values:
                    lite (default): returns key metadata for the given search terms
                    idlist: returns a list of IDs and sources for the given search terms
                    core: returns full metadata for a given publication ID; including
                        abstract, full text links, and MeSH terms
                pageSize (int): defaults to 25 publications per page; maximum allowed is 1000

        Returns:
        """
        self._check_query(**kwargs)
        body = self.common_params | kwargs
        endpoint = f"{self.base_url}searchPOST/"
        headers = {
            "Content-Type": "application/x-www-form-urlencoded",
        }

        return self.session.post(endpoint, data=body, headers=headers)

    @staticmethod
    def escape_brackets(search_term: str) -> str:
        """
        Brackets have special meaning in EuropePMC queries. Thus, they must be escaped
        if they are part of search terms [for example, in the doi 10.1016/S2666-5247(22)00181-1]
        Args:
            search_term: the search term that might contain brackets to be escaped

        Returns: search_term with brackets escaped

        """
        return search_term.replace("(", "\(").replace(")", "\)")

    def query_doi(self, doi: str, **kwargs) -> dict:
        """
        Searches EuropePMC for publications whose DOI matches the doi argument.
        If zero or more than one result is found, raises an error

        Args:
            doi: A DOI string such as 10.1093/bioinformatics/btac311

        Returns: Dictionary representation of the publication matching the queried DOI
        """
        r = self.search(query=f"doi:{self.escape_brackets(doi)}", **kwargs)
        if (hit_count := r["hitCount"]) == 0:
            self.logger.info(f"Query for DOI {doi} did not return any results")
            return dict()
        else:
            assert 1 == hit_count, f"Query for DOI {doi} returned multiple results"
            return r["resultList"]["result"][0]

    def query_doi_list(self, **kwargs) -> dict:
        """
        Args:
            **kwargs:
                 doi_list (required, list): DOIs to query
                 any other kwargs to be passed to search_post (see documentation for that method)
        Returns:

        """
        doi_list = kwargs.pop("doi_list")
        page_size = kwargs.pop("pageSize", None)
        if page_size is None:
            page_size = 1000
        list_length = len(doi_list)
        escaped_list = [self.escape_brackets(doi) for doi in doi_list]
        assert page_size > list_length, (
            f"This method does not support pagination yet and "
            f"the number of DOIs queried ({list_length}) exceeds"
            f"pageSize ({page_size}). Either increase pageSize if possible"
            f"(server-side limit is 1000) or make more than one call to this method"
        )
        query = f"doi:({' OR '.join(escaped_list)})"
        response = self.search_post(query=query, pageSize=page_size, **kwargs)

        if (hit_count := response["hitCount"]) < list_length:
            self.logger.warning(
                "List of publications returned by EuropePMC shorter than queried DOI list",
                extra={"epmc_results": hit_count, "queried_dois": list_length},
            )

        doi_dict = dict()
        for r in response["resultList"]["result"]:
            try:
                doi_dict[r["doi"]] = r
            except KeyError:
                self.logger.error("No doi field!", extra=r)

        return doi_dict
